#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "hmp2mid.h"
#include "cfile.h"
#include "args.h"
#include "digi.h"

enum { NO_MUSIC, TIMIDITY, PLAYMUS };

/* Use playmus by default because playmus is much CPU friendlier. Timidity has
   the best sound, but timidity's CPU-usage is too much, especially when
   (re)starting a song. This makes the game all stuttery. */
static int midi_driver = PLAYMUS;

/* Volumes are 0-127, timidity's volume is in % but allows volume > 100%
   and is a bit soft by default */
static int timidity_volume;
static int midi_volume = 127;
static void digi_stop_current_song();

FILE *midiplayer_stdin = NULL;

void digi_play_midi_song(char * filename, char *melodic_bank, char *drum_bank,
  int loop)
{
  static int first_time = 1;
  char *p, mid_filename[512], command[512];

  if (first_time)
  {
    if(FindArg("-nomusic"))
      midi_driver = NO_MUSIC;

    if(FindArg("-timidity"))
      midi_driver = TIMIDITY;

    if(FindArg("-playmus"))
      midi_driver = PLAYMUS;
    
    atexit(digi_stop_current_song);
    first_time = 0;
  }

  if (midi_driver == NO_MUSIC)  
    return;
  
  /* Stop current playing song */
  digi_stop_current_song();
  if (!filename)
    return;
  
  /* We need to write the converted hmp to a file, to avoid tmpfile problems
     we store the converted files under ~/.d1x, where we can use them as
     a cache too, speeding things up a bit. */
  p = getenv("HOME");
  snprintf(mid_filename, sizeof(mid_filename), "%s/.d1x/%s", p? p:".",
    filename);
  if ((p = strstr(mid_filename, ".hmp"))) strcpy(p, ".mid");
  if (access(mid_filename, R_OK) != 0)
  {
    const char *err;
    CFILE *hmp_in;
    FILE *mid_out = fopen(mid_filename, "w");
    if (!mid_out)
    {
      fprintf(stderr, "Error could not open: %s for writing: %s\n",
        mid_filename, strerror(errno));
      return;
    }
    hmp_in = cfopen(filename, "rb");
    if (!hmp_in)
    {
      fprintf(stderr, "Error could not open: %s\n", filename);
      fclose(mid_out);
      return;
    }
    err = hmp2mid((hmp2mid_read_func_t)cfread, hmp_in, mid_out);
    fclose(mid_out);
    cfclose(hmp_in);
    if (err)
    {
      fprintf(stderr, "%s\n", err);
      unlink(mid_filename);
      return;
    }
  }
  
  /* And execute our midiplayer with the found or created midi file */
  switch (midi_driver)
  {
    case TIMIDITY:
      snprintf(command, sizeof(command),
        "timidity -in%s -A %d %s > /dev/null",
          loop? "l":"", midi_volume, mid_filename);
      timidity_volume = midi_volume;
      break;
    case PLAYMUS:
      snprintf(command, sizeof(command),
        "playmus -i%s -r 44100 -v %d %s > /dev/null",
          loop? " -l":"", midi_volume, mid_filename);
      break;
  }
  midiplayer_stdin = popen(command, "w");
  if (!midiplayer_stdin)
    fprintf(stderr, "Error couldn't execute: %s\n", command);
}

void digi_stop_current_song()
{
  if (midiplayer_stdin)
  {
    switch (midi_driver)
    {
      case TIMIDITY:
        fputc('q', midiplayer_stdin);
        break;
      case PLAYMUS:
        fputs("h\r", midiplayer_stdin);
        break;
    }
    fflush(midiplayer_stdin);
    pclose(midiplayer_stdin);
    midiplayer_stdin = NULL;
  }
}

void digi_set_midi_volume(int mvolume)
{
  midi_volume = mvolume;
  if (midiplayer_stdin)
  {
    switch (midi_driver)
    {
      case TIMIDITY:
        while((timidity_volume + 5) < midi_volume)
        {
          fputc('V', midiplayer_stdin);
          fflush(midiplayer_stdin);
          timidity_volume += 10;
        }
        while((timidity_volume - 5) > midi_volume)
        {
          fputc('v', midiplayer_stdin);
          fflush(midiplayer_stdin);
          timidity_volume -= 10;
        }
        break;
      case PLAYMUS:
        fprintf(midiplayer_stdin, "v%d\r", midi_volume);
        fflush(midiplayer_stdin);
        break;
    }
  }
}
